#include <stdlib.h>
#include <stdio.h>

/* Definizione */
struct listNode {
	int data;
	struct listNode *next;
};

/* Prototipi funzioni */
void insert_head(struct listNode **, int);
void append(struct listNode **, int);
int delete_head(struct listNode **);
void print_list(struct listNode *);
void insert_sorted(struct listNode **, int);
int delete(struct listNode **, int);
void free_memory(struct listNode *);

/* Inserire un nodo in testa alla lista */
void insert_head(struct listNode **head_ptr, int new_value)
{
    struct listNode *new_node;	/* Puntatore al nuovo nodo da inserire */

    new_node = malloc(sizeof(struct listNode));
	if (new_node == NULL) {
		printf("Impossibile allocare memoria!\n");
		exit(EXIT_FAILURE);
	}

	new_node->data = new_value;
	/* Inserimento in testa alla lista */
	new_node->next = *head_ptr;
	*head_ptr = new_node; 
}

/* Appendere (inserire in coda) un nodo alla lista */
void append(struct listNode **head_ptr, int new_value)
{
    struct listNode *current;    /* Puntatore al nodo corrente */
    struct listNode *new_node;    /* Puntatore al nuovo nodo da appendere */
    
    new_node = malloc(sizeof(struct listNode));
    if (new_node == NULL) {
        printf("Impossibile allocare memoria!\n");
        exit(EXIT_FAILURE);
    }
    new_node->data = new_value;
    new_node->next = NULL;
    
    current = *head_ptr;
    if (current == NULL) {    /* Lista vuota */
        *head_ptr = new_node;    /* Aggiornamento del puntatore alla testa della lista */
        return;
    }
    
    /* Scansione della lista fino all'ultimo nodo */
    while (current->next != NULL)
        current = current->next;
    
    current->next = new_node;    /* Inserimento del nuovo nodo */
}

/* Cancellare il nodo in testa alla lista */
int delete_head(struct listNode **head_ptr)
{
    struct listNode *temp;
    
    if (*head_ptr == NULL) /* Lista vuota */
        return 0;
    
    temp = *head_ptr; /* Salva il puntatore al primo nodo da rimuovere */
    
    /* Cancellazione del primo nodo */
    *head_ptr = temp->next;
    free(temp);
    return 1;
}

void print_list(struct listNode *head)
{
    struct listNode *current;
    
    if (head == NULL) /* La lista è vuota */
        printf("La lista è vuota\n\n");
    else {
        printf("La lista è:\n");
        
        /* Stampa nodo per nodo fino alla fine della lista */
        for (current=head; current != NULL; current=current->next)
            printf("%d --> ", current->data);
        printf("NULL\n\n");
    }
}

/* Inserire un nodo in modo ordinato in base al valore */
void insert_sorted(struct listNode **head_ptr, int new_value)
{
    struct listNode *current; /* Puntatore al nodo corrente */
    struct listNode *new_node; /* Puntatore al nuovo nodo da inserire */

    new_node = malloc(sizeof(struct listNode));
	if (new_node == NULL) {
		printf("Impossibile allocare memoria!\n");
		exit(EXIT_FAILURE);
	}
	new_node->data = new_value;

	current = *head_ptr;
	if (current == NULL) { //La lista è vuota		
		new_node->next = *head_ptr;
		*head_ptr = new_node;
		return;
	}

	/* Inserimento in modo ordinato */
	while (current->next != NULL && current->next->data < new_value)
		current = current->next;

    if (*head_ptr == current && current->data > new_value) { //Elemento da inserire in testa
        new_node->next = current;
        *head_ptr = new_node;
    }
    else {
        new_node->next = current->next;
        current->next = new_node;
    }
}

/* Cancellare un nodo della lista */
int delete(struct listNode **head_ptr, int value)
{
    struct listNode *current;	/* Puntatore al nodo corrente */
    struct listNode *previous;	/* Puntatore al nodo precedente */

	if (*head_ptr == NULL) /* Lista vuota */
		return 0;

	current = *head_ptr;

	if (current->data == value) { /* Cancellazione del primo nodo */
		*head_ptr = current->next; /* Aggiorna il puntatore alla testa */
		free(current);
		return 1;
	}
    
    /* Scansione della lista a partire dal secondo nodo */
	previous = current;
	current = current->next;
	while (current != NULL) {
		if (current->data == value) {
			previous->next = current->next; /* Aggiornamento del puntatore al nodo successivo nel nodo che precede quello da cancellare */
			free(current);
			return 1;
		}
        previous = current;
		current = current->next;
	}
    return 0;
}

void free_memory(struct listNode *head)
{
	while (head != NULL) {
		struct listNode *temp = head;
		/* NOTE: cannot free 'head' here! */

		head = head->next;
		free(temp); 
	}
}

int main()
{
	/* Definisce una nuova lista vuota. */
	struct listNode *list_head = NULL;
    int res;

	print_list(list_head);
    
	append(&list_head, 1);
	print_list(list_head);
	
    append(&list_head, 5);
    print_list(list_head);
    
	insert_head(&list_head, 0);
	print_list(list_head);
    
	insert_sorted(&list_head, 3);
	print_list(list_head);

    insert_sorted(&list_head, -1);
    print_list(list_head);

    insert_sorted(&list_head, 1);
    print_list(list_head);

    res = delete_head(&list_head);
    if (!res)
        printf("Lista vuota.\n\n");
	print_list(list_head);
	
    res = delete(&list_head, 2);
    if (!res)
        printf("Elemento non trovato.\n");
	print_list(list_head);
    
	res = delete(&list_head, 1);
    if (!res)
        printf("Elemento non trovato.\n");
	print_list(list_head);

    res = delete(&list_head, 3);
    if (!res)
        printf("Elemento non trovato.\n");
    print_list(list_head);

	free_memory(list_head);

	return 0;
}
