/* Inserting and deleting nodes in a list 

   Note: if insertion in a sorted list and insertion 
   at the head of the list are both used, the order of the 
   elements is no longer preserved */

#include <stdio.h>
#include <stdlib.h>

typedef char dataType;    /* defined by developer */

/* Definition */
struct listNode { 
  dataType data; 
  struct listNode *next; 
};
typedef struct listNode ListNode; /* synonim for struct listNode */
typedef ListNode *List;	   /* synonim for ListNode*  */


/* Equivalent definitions */
/* typedef struct listNode { 
  dataType data;
  struct listNode *next; 
} ListNode;
typedef ListNode *List; */

/* typedef struct ListNode *List;
typedef struct listNode { 
  dataType data;
  List next; 
} ListNode; */

/* typedef struct listNode *List;
typedef struct listNode { 
  dataType data;
  List next; 
} ListNode; */

/* Prototypes */
void insertSorted(List *, char);
void insertHead(List *, char);
char delete(List *, char);
char deleteHead(List *);
int isEmpty(List);
void printList(List);
void instructions(void);

int main(void)
{
   List head = NULL;  /* initially there are no nodes */
   unsigned int choice=0;        /* user's choice */
   char item;         /* char entered by user */

   instructions();    /* display the menu */
   printf("? ");
   while (scanf("%d", &choice) != 1) {
      printf("Please enter an integer to choose: ");
      scanf("%*s"); /* Flag character * tells scanf to read and ignore the value */
   }
 
   /* Loop while user does not choose 5 */
   while (choice != 5) {
      switch (choice) {
         case 1:  /* insert item at the head of the list */
            printf("Enter a character: ");
            scanf("\n%c", &item);
            insertHead(&head, item); 
            printList(head);
            break;
         case 2:  /* insert item in the sorted list */
            printf("Enter a character: ");
            scanf("\n%c", &item);
            insertSorted(&head, item); 
            printList(head);
            break;
         case 3:  /* delete first element */
            item = deleteHead(&head);
            printf("%c deleted.\n", item);
            printList(head);
            break;
         case 4:  /* delete an item */
            /* if list is not empty */
            if (!isEmpty(head)) {
               printf("Enter character to be deleted: ");
               scanf("\n%c", &item);
            /* if item is found, remove it */
               if (delete(&head, item)) {
                  printf("%c deleted.\n", item);
                  printList(head);
               }
               else
                  printf("%c not found.\n\n", item);
            }
            else
               printf("List is empty.\n\n"); 
            break;
         default:
            printf("Invalid choice.\n\n");
            instructions();
            break;
      } /* end switch */

      printf("? ");
      while (scanf("%d", &choice) != 1) {
         printf("Please enter an integer to choose: ");
         scanf("%*s"); /* Flag character * tells scanf to read and ignore the value */
      }
   } /* end while choice */

   printf("End of run.\n");
   return 0;
}

/* Print the instructions */
void instructions(void)
{
   printf("Enter your choice:\n"
          "   1. Insert an element at the head of the list.\n"
          "   2. Insert an element in the sorted list.\n"
          "   3. Delete the first element from the list.\n"
          "   4. Delete an element from the list.\n"
          "   5. Quit\n");
}

/* Insert new node at the beginning of the list */
void insertHead(List *s, char value)
{
   List new;    /* Pointer to new node */

   new = malloc(sizeof(ListNode));  /* Create node */

   if (new != NULL) {     /* Is space available? */
      new->data = value;  /* Place value in node */
      new->next = *s;
      *s = new;
   }
   else
      printf("%c not inserted. No memory available.\n", value);
}


/* Insert a new value into the list in sorted order */
void insertSorted(List *s, char value)
{
   List new;      /* Pointer to new node */
   List previous; /* Pointer to previous node in list */
   List current;  /* Pointer to current node in list */

   new = malloc(sizeof(ListNode));  /* Create node */

   if (new != NULL) {    /* Is space available? */
      new->data = value;
      new->next = NULL;

      previous = NULL;
      current = *s;

      /* Loop to find the correct location in the list */
      while (current != NULL && value > current->data) {
         previous = current;          /* walk to ...   */
         current = current->next;  /* ... next node */
      }

      /* Insert new node at beginning of list */
      if (previous == NULL) {
         new->next = *s;
         *s = new;
      }
      else {  /* Insert new node between previousPtr and currentPtr */
         previous->next = new;
         new->next = current;
      }
   }
   else
      printf("%c not inserted. No memory available.\n", value);
}

/* Delete the first node of the list and return the deleted value */
char deleteHead(List *s)
{
   List temp;     /* Temporary node pointer */
   char value;

   temp = *s;

   if (temp != NULL) {
     *s = temp->next;
     value = temp->data;
     free(temp);
     return value;
   }
   return '\0';
}

/* Delete a list element and return the deleted value */
char delete(List *s, char value)
{
   List previous; /* Pointer to previous node in list */
   List current;  /* Pointer to current node in list */
   List temp;     /* Temporary node pointer */

   /* Delete first node */
   if ((*s)->data == value) {
      temp = *s;        /* Hold onto node being removed */
      *s = (*s)->next;  /* De-thread the node */
      free(temp);       /* Free the de-threaded node */
      return value;
   }
   else {
      previous = *s;
      current = (*s)->next;

      /* Loop to find the correct location in the list */
      while (current != NULL && current->data != value) {
         previous = current;          /* walk to ...   */
         current = current->next;  /* ... next node */
      }

      /* Delete node at currentPtr */
      if (current != NULL) {
         temp = current;
         previous->next = current->next;
         free(temp);
         return value;
      }                                                        
   }

   return '\0';
}

/* Return 1 if the list is empty, 0 otherwise */
int isEmpty(List s)
{
   return s == NULL;
}

/* Print the list */
void printList(List current)
{
   /* If list is empty */
   if (current == NULL)
      printf("List is empty.\n\n");
   else {
      printf("The list is:\n");
 
      /* While not the end of the list */
      while (current != NULL) {
         printf("%c --> ", current->data);
         current = current->next;
      }

      printf("NULL\n\n");
   }
}
